from yta_validation.parameter import ParameterValidator
from yta_temp import Temp
from yta_logger import ConsolePrinter
from pathlib import Path
from typing import Union

import av
import subprocess


CACHE_PREFIX: str = 'cache_yta'
"""
The prefix we use for the temporary files we create
by using the cache.
"""

class FramesCacheFactory:
    """
    Factory to simplify the access to our cached frames.

    Use its static methods to obtain the frames.
    """

    @staticmethod
    def black_video_frame(
        size: tuple[int, int],
        video_fps: float,
        transparency: Union[float, None] = None
    ):
        """
        Get a black video frame by using the cache system. The
        frame will be transparent if the `transparency` value
        provided is a float (between 0.0 and 1.0), or opaque 
        if it is None.
        """
        return BlackFrameCache.get(
            size = size,
            video_fps = video_fps,
            transparency = transparency
        )

    @staticmethod
    def silent_audio_frame(
        audio_fps: float,
        audio_layout: str,
        audio_format: str,
        audio_samples_per_frame: float
    ):
        """
        Get a silent audio frame by using the cache system.
        """
        return SilentAudioFrameCache.get(
            audio_fps = audio_fps,
            audio_layout = audio_layout,
            audio_format = audio_format,
            audio_samples_per_frame = audio_samples_per_frame
        )

class BlackFrameCache:
    """
    A black video frames cached generator, useful when we
    need the same black frames again and again.

    This class will create temporary files (in the WIP folder)
    to load them directly with the pyav library.
    """

    _cache = {}
    """
    The cache that is able to store all the black frames that
    we create during the process. The frame specifications 
    will be the key to access it in the dict.
    """

    @staticmethod
    def get(
        size: tuple[int, int],
        video_fps: float,
        transparency: Union[float, None] = None
    ) -> 'VideoFrame':
        """
        Get a black video frame (transparent or not, according
        to the `transparency` parameter provided) of the given
        `size` and with the also provided `video_fps` and
        `pixel_format`.

        The video frames will be cached so the access the
        next time is instantaneous.

        TODO: This seems to be unnecessary as we can generate
        the frame in memory and we don't need to write it...
        """
        ParameterValidator.validate_number_between('transparency', transparency, 0.0, 1.0)
        
        pixel_format = (
            'rgb24'
            if transparency is None else
            'rgba'
        )

        key = (size[0], size[1], video_fps, pixel_format)

        if key in BlackFrameCache._cache:
            return BlackFrameCache._cache[key]

        width, height = size

        # Temporary specific file
        filename = (
            f'{CACHE_PREFIX}_black_{width}x{height}_{video_fps}_{pixel_format}.mp4'
            if transparency is None else
            f'{CACHE_PREFIX}_transparent_{str(transparency)}_black_{width}x{height}_{video_fps}_{pixel_format}.mp4'
        )
        filename = Temp.get_custom_wip_filename(filename)

        # 'black@0.0' means fully transparent, so we transform
        # our parameter to fit the str we need
        transparency = 1.0 - transparency

        color = (
            'black'
            if transparency is None else
            # Color must be something like 'black@0.0'
            f'black@{str(transparency)}'
        )

        if not Path(filename).exists():
            cmd = [
                'ffmpeg',
                '-y',
                '-f',
                'lavfi',
                '-i',
                f'color=size={width}x{height}:color={color}',
                '-pix_fmt',
                pixel_format,
                '-frames:v',
                '1',
                '-r',
                str(video_fps),
                str(filename)
            ]
            # TODO: We shouldn't do this with a subprocess... it
            # is not a good thing when we want APIs...
            subprocess.run(
                cmd,
                stdout = subprocess.DEVNULL,
                stderr = subprocess.DEVNULL
            )

            ConsolePrinter().print(f'The black cached (transparency={str(transparency)}) frame has been created as "{filename}"')

        # Load decoded frame
        container = av.open(str(filename))
        frame = next(container.decode(video = 0))

        # Store in cache
        BlackFrameCache._cache[key]: 'VideoFrame' = frame

        return frame

class SilentAudioFrameCache:
    """
    A silent audio frames cached generator, useful when we
    need the same silent frames again and again.

    This class will create temporary files (in the WIP folder)
    to load them directly with the pyav library.
    """

    _cache = {}
    """
    The cache that is able to store all the silent frames that
    we create during the process. The frame specifications 
    will be the key to access it in the dict.
    """

    @staticmethod
    def get(
        # TODO: Review the parameter types and that...
        audio_fps: float,
        audio_layout: str,
        audio_format: str,
        audio_samples_per_frame: float
    ):
        """
        Get a silent audio frame with the provided `audio_fps`,
        `audio_layout`, `audio_format` and `audio_samples_per_frame`.

        The audio frames will be cached so the access the
        next time is instantaneous.
        """
        key = (audio_fps, audio_layout, audio_format, audio_samples_per_frame)

        if key in SilentAudioFrameCache._cache:
            return SilentAudioFrameCache._cache[key]

        # Temporary specific file
        filename = f'{CACHE_PREFIX}_silence_{audio_fps}_{audio_layout}_{audio_format}_{audio_samples_per_frame}.wav'
        filename = Temp.get_custom_wip_filename(filename)

        if not Path(filename).exists():
            duration = audio_samples_per_frame / audio_fps

            cmd = [
                'ffmpeg', '-y',
                '-f', 'lavfi',
                '-i', f'anullsrc=channel_layout={audio_layout}:sample_rate={audio_fps}',
                '-t', str(duration),
                #'-c:a', audio_format,
                # TODO: What about this forced format (?)
                '-c:a', 'pcm_f32le',
                # Apparently this '-frames:a' is stupid and we need the '-t' instead
                #'-frames:a', '1',
                str(filename)
            ]
            # TODO: We should do this with a subprocess... it is
            # not a good thing when we want to create services
            subprocess.run(cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL)

            ConsolePrinter().print(f'The silent audio cached frame has been created as "{filename}"')

        # Load decoded frame
        container = av.open(str(filename))
        frame = next(container.decode(audio = 0))

        # Store in cache
        SilentAudioFrameCache._cache[key]: 'AudioFrame' = frame

        return frame



from yta_editor.utils.frame_wrapper import AudioFrameWrapped
from yta_editor.utils import VideoUtils
from yta_constants.ffmpeg import FfmpegAudioLayout, FfmpegAudioFormat
from yta_video_frame_time.t_fraction import fps_to_time_base
from quicktions import Fraction
from typing import Union


# TODO: Is this method here ok (?)
def generate_silent_frames(
    video_fps: float,
    audio_fps: int,
    audio_samples_per_frame: int,
    layout: str = FfmpegAudioLayout.STEREO.value,
    format: str = FfmpegAudioFormat.FLTP.value,
    is_from_gap: bool = False
) -> list[AudioFrameWrapped]:
    """
    Get the audio silent frames we need for
    a video with the given `video_fps`,
    `audio_fps` and `audio_samples_per_frame`,
    using the also provided `layout` and
    `format` for the audio frames.

    This method is used when we have empty
    items on our tracks and we need to 
    provide the frames, that are passed as
    AudioFrameWrapped instances and tagged as
    coming from empty items.
    """
    # Check how many full and partial silent
    # audio frames we need
    number_of_frames, number_of_remaining_samples = _audio_frames_and_remainder_per_video_frame(
        video_fps = video_fps,
        sample_rate = audio_fps,
        number_of_samples_per_audio_frame = audio_samples_per_frame
    )
    
    frames = (
        [
            AudioFrameWrapped(
                # This could be created once and then copied
                frame = VideoUtils.audio.frame.silent(
                    sample_rate = audio_fps,
                    number_of_samples = audio_samples_per_frame,
                    layout = layout,
                    format = format,
                    pts = None,
                    time_base = None
                ),
                is_from_gap = is_from_gap,
                is_silent = True
            )
        ] * number_of_frames
        if number_of_frames > 0 else
        []
    )

    # The remaining partial silent frames we need
    if number_of_remaining_samples > 0:
        silent_frame = VideoUtils.audio.frame.silent(
            sample_rate = audio_fps,
            number_of_samples = number_of_remaining_samples,
            # TODO: Check where do we get this value from
            layout = layout,
            # TODO: Check where do we get this value from
            format = format,
            pts = None,
            time_base = None
        )
        
        frames.append(
            AudioFrameWrapped(
                frame = silent_frame,
                is_from_gap = is_from_gap,
                is_silent = True
            )
        )

    return frames

def _audio_frames_and_remainder_per_video_frame(
    # TODO: Maybe force 'fps' as int (?)
    video_fps: Union[float, Fraction],
    sample_rate: int, # audio_fps
    number_of_samples_per_audio_frame: int
) -> tuple[int, int]:
    """
    *For internal use only*

    Get how many full silent audio frames we
    need and the remainder for the last one
    (that could be not complete), according
    to the parameters provided.

    This method returns a tuple containing
    the number of full silent audio frames
    we need and the number of samples we need
    in the last non-full audio frame.
    """
    # Video frame duration (in seconds)
    time_base = fps_to_time_base(video_fps)
    sample_rate = Fraction(int(sample_rate), 1)

    # Example:
    # 44_100 / 60 = 735  ->  This means that we
    # will have 735 samples of sound per each
    # video frame
    # The amount of samples per frame is actually
    # the amount of samples we need, because we
    # are generating it...
    samples_per_frame = sample_rate * time_base
    # The 'nb_samples' is the amount of samples
    # we are including on each audio frame
    full_audio_frames_needed = samples_per_frame // number_of_samples_per_audio_frame
    remainder = samples_per_frame % number_of_samples_per_audio_frame
    
    return int(full_audio_frames_needed), int(remainder)


from yta_editor.media.video import VideoFileMedia
from yta_editor.transformations.effects.abstract import EffectParameters
from yta_editor.transformations.effects.video import VideoEffect
from yta_editor_nodes.processor.video.alphapedia_mask_reveal import AlphaPediaMaskRevealVideoNodeProcessor
from yta_editor_time.evaluation_context import EvaluationContext
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union
from dataclasses import dataclass


# TODO: We need to define all the specific effect
# params to be able to build them properly missing
# no fields
@dataclass
class AlphaPediaMaskRevealVideoEffectParameters(EffectParameters):
    """
    The parameters to use when applying this effect
    that reveals a video with an alpha pedia mask for
    a specific `evaluation_context`.

    This will be returned by the effect when
    calculated, for a specific `evaluation_context`.
    """

    def __init__(
        self,
        mask_frame: Union['np.ndarray', 'moderngl.Texture']
    ):
        # TODO: Add doc
        self.mask_frame: Union['np.ndarray', 'moderngl.Texture'] = mask_frame
        """
        The frame that must be used to be applied as
        the mask frame.
        """
        

class AlphaPediaMaskRevealVideoEffect(VideoEffect):
    """
    The effect that will modify the video by
    revealing it with an alpha video mask according
    to the provided conditions.
    """

    def __init__(
        self,
        mask_input_filename: str,
        do_use_gpu: bool = True
    ):
        self.do_use_gpu: bool = do_use_gpu
        """
        Flag to indicate if using GPU or not.
        """
        self._mask_video: VideoFileMedia = VideoFileMedia(
            filename = mask_input_filename
        )
        """
        *For internal use only*

        The alpha video that we will use to read the
        frames and apply the mask.
        """
        self._node_processor = AlphaPediaMaskRevealVideoNodeProcessor(
            # TODO: What do I do with the 'opengl_context' (?)
            opengl_context = None
        )
        """
        *For internal use only*

        The node processor that we will use to process
        the input frame with the corresponding mask
        frames.
        """

    def _get_params_at(
        self,
        evaluation_context: EvaluationContext
    ) -> AlphaPediaMaskRevealVideoEffectParameters:
        """
        *For internal use only*

        Get the parameters that must be applied at the given
        `evaluation_context`.
        """
        ParameterValidator.validate_mandatory_instance_of('evaluation_context', evaluation_context, EvaluationContext)

        # TODO: This is repeated in the 'transition_calculator.py'
        # TODO: What about an easing function (?)
        t_transition_media = evaluation_context.progress.value * (self._mask_video.number_of_frames - 1) / self._mask_video.fps

        video_frame = self._mask_video.get_video_frame_at(
            t_media = t_transition_media
        ).to_ndarray(format = 'rgba')
        
        video_frame = (
            video_frame.frame
            if PythonValidator.is_instance_of(video_frame, 'VideoFrame') else
            video_frame
        )

        return AlphaPediaMaskRevealVideoEffectParameters(
            mask_frame = video_frame
        )
    
    def _apply(
        self,
        # TODO: Set the type
        frame: any,
        output_size: Union[tuple[int, int], None],
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> any:
        parameters = self._get_params_at(
            evaluation_context = evaluation_context
        )

        return self._node_processor.process(
            inputs = {
                'base_input': frame,
                'mask_input': parameters.mask_frame
            },
            evaluation_context = evaluation_context,
            do_use_gpu = self.do_use_gpu,
            output_size = output_size
        )


"""
TODO: Check if this class can be refactored, moved
to somewhere else or put in other library.
"""
from yta_constants.enum import YTAEnum as Enum
import numpy as np


class FrameAdapterMode(Enum):
    """
    The different modes we have to adapt the frames to
    a specific resolution.
    """

    ALPHA_BACKGROUND = 'alpha_background'
    """
    A full transparent frame will be created with the
    expected resolution and the original frame will be
    placed in the middle.
    """

class FrameAdapter:
    """
    A class to modify the frames and adapt them to the 
    resolution we need for the timeline. This is needed
    to be able to use them as inputs in the nodes.
    """

    @staticmethod
    def adapt(
        frame: np.ndarray,
        resolution: tuple[int, int] = (1920, 1080),
        mode: FrameAdapterMode = FrameAdapterMode.ALPHA_BACKGROUND
    ) -> np.ndarray:
        """
        Adapt the `frame` provided to fit the expected and
        also given `resolution` by using the `adapt_mode`
        provided as parameter.
        """
        mode = FrameAdapterMode.to_enum(mode)

        if frame.size == resolution:
            return frame

        # TODO: Improve this, please
        if mode == FrameAdapterMode.ALPHA_BACKGROUND:
            return _adapt_with_alpha(
                frame = frame,
                resolution = resolution
            )

# Specific methods below
def _adapt_with_alpha(
    frame: np.ndarray,
    resolution: tuple[int, int]
) -> np.ndarray:
    """
    Creates a fully alpha background with the `resolution`
    given and puts the `frame` provided in the middle of
    it.

    This is the functionality to be done when using the
    `ALPHA_BACKGROUND` option.
    """
    if frame.ndim != 3:
        raise Exception('The frame must be HxWxC.')

    src_h, src_w, src_c = frame.shape

    if src_c not in (3, 4):
        raise Exception('The frame must be RGB or GBA.')

    # Final transparent frame
    dst = np.zeros((resolution[1], resolution[0], 4), dtype = frame.dtype)

    # RGBA if necessary
    if src_c == 3:
        src_rgba = np.empty((src_h, src_w, 4), dtype = frame.dtype)
        src_rgba[..., :3] = frame
        src_rgba[..., 3] = np.iinfo(frame.dtype).max
    else:
        src_rgba = frame

    # Offsets to center
    offset_x = (resolution[0] - src_w) // 2
    offset_y = (resolution[1] - src_h) // 2

    if (
        offset_x < 0 or
        offset_y < 0
    ):
        # TODO: Do rescale and crop if needed
        raise Exception('Original frame is bigger than the resolution requested.')

    # Copy it
    dst[
        offset_y:offset_y + src_h,
        offset_x:offset_x + src_w
    ] = src_rgba

    return dst

# TODO: Create the GPU node that does this



from yta_editor.media.abstract import _Media
from yta_editor.sources.abstract import _AudioSource
from yta_editor.sources.audio import AudioFileSource, AudioNumpySource
from yta_logger import ConsolePrinter
# from yta_editor.utils.effects import apply_audio_effects_to_frame_at
from yta_video_pyav.reader.filter.dataclass import GraphFilter
from quicktions import Fraction
from typing import Union

import copy


class _AudioMedia(_Media):
    """
    Abstract class to be inherited by any
    media element.

    The media element is an element that
    includes a source and a 't_start' and
    't_end' values to be able to subclip that
    media source and use only the part we
    want to use.
    """

    def __init__(
        self,
        source: _AudioSource,
        t_start: Union[int, float, Fraction] = 0.0,
        t_end: Union[int, float, Fraction, None] = None,
        audio_effects: list['SerialNode', 'ParallelNode'] = []
    ):
        super().__init__(
            source = source,
            t_start = t_start,
            t_end = t_end,
            audio_effects = audio_effects
        )

    def get_audio_frames_with_t_source(
        self,
        t_source: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction]
    ):
        """
        Get the sequence of audio frames for a 
        given video 't_source' time moment, using
        the audio cache system.

        The `t_source` time moment must be a value in
        the time interval `[0, source.t_end]`, that will
        be used directly to be read from the source.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        from yta_validation.parameter import ParameterValidator

        ParameterValidator.validate_mandatory_number_between('t_source', t_source, 0, self.source.duration, True, False)

        # ConsolePrinter().print(f'Getting audio frames from "t_source={str(float(t_source))}"')
        
        for frame in self.source.get_audio_frames_at(t_source, video_fps):
            # TODO: Effects must be now applied with nodes
            # yield apply_audio_effects_to_frame_at(
            #     effects_stack = self._audio_effects,
            #     frame = frame,
            #     t = t
            # )
            yield frame

    # @with_t_adjusted_to_source
    def get_audio_frames_at(
        self,
        t_media: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ):
        """
        Get the sequence of audio frames for a 
        given video 't_media' time moment, using
        the audio cache system.

        The `t_media` time moment must be a value
        between 0 and the media duration, that will
        be transformed into the corresponding source
        time moment to be read.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        # TODO: Validate (?)
        t_source = self._t_media_to_t_source(
            t_media = t_media,
            do_use_source_limits = do_use_source_limits
        )

        ConsolePrinter().print(f'Getting audio frames from "t_media={str(float(t_media))}" that is actually "t_source={str(float(t_source))}"')
        for frame in self.source.get_audio_frames_at(t_source, video_fps):
            # TODO: Effects must be now applied with nodes
            # yield apply_audio_effects_to_frame_at(
            #     effects_stack = self._audio_effects,
            #     frame = frame,
            #     t = t
            # )
            yield frame

class AudioFileMedia(_AudioMedia):
    """
    An audio media that is read from an audio
    file and can be subclipped to a specific
    time range.
    """

    @property
    def copy(
        self
    ) -> 'AudioFileMedia':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        instance_copy = AudioFileMedia._init_with_source(
            source = self.source,
            t_start = self.t_start,
            t_end = self.t_end,
            audio_effects = copy.deepcopy(self._audio_effects)
        )

        return instance_copy

    @property
    def audio_fps(
        self
    ) -> Union[int, None]:
        """
        The frames per second of the audio.
        """
        return self.source.audio_fps
    
    @property
    def audio_codec_name(
        self
    ) -> Union[str, None]:
        """
        The name of the audio codec.
        """
        return self.source.audio_codec_name
    
    @property
    def audio_layout(
        self
    ) -> Union[str, None]:
        """
        The audio layout.
        """
        return self.source.audio_layout.name
    
    @property
    def audio_format(
        self
    ) -> Union[str, None]:
        """
        The audio format.
        """
        return self.source.audio_format.name
    
    @property
    def audio_time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the audio.
        """
        return self.source.audio_time_base
    
    def __init__(
        self,
        filename: str,
        t_start: Union[int, float, Fraction] = 0.0,
        t_end: Union[int, float, Fraction, None] = None,
        # These are ffmpeg filters
        audio_effects: list[Union['SerialNode', 'ParallelNode']] = [],
        audio_filters: list[GraphFilter] = []
    ):
        # We first create the source with no
        # filters just to be able to know the
        # original size and properties
        source = AudioFileSource(
            filename = filename,
            audio_filters = audio_filters
        )

        t_end = (
            source.duration
            if t_end is None else
            t_end
        )

        super().__init__(
            source = AudioFileSource(
                filename = filename,
                audio_filters = audio_filters
            ),
            t_start = t_start,
            t_end = t_end,
            audio_effects = audio_effects
        )

    def add_audio_filter(
        self,
        filter: GraphFilter
    ) -> 'AudioFileMedia':
        """
        Add an audio filter to the list of filters
        to apply.
        """
        self.source.add_audio_filter(filter)

        return self
    
    def set_audio_filters(
        self,
        filters: list[GraphFilter] = []
    ) -> 'AudioFileMedia':
        """
        Set the 'filters' provided as the new audio
        filters, replacing the previous one if 
        existing.
        """
        # TODO: What if 'video' filters coming (?)
        self.source.set_audio_filters(filters)

        return self

# TODO: Create 'AudioNumpyMedia'



from yta_editor.transformations.effects.abstract import EffectParameters
from yta_editor.transformations.effects.video import VideoEffect
from yta_editor_nodes.processor.video.waving_frame import WavingFrameVideoNodeProcessor
from yta_editor_time.evaluation_context import EvaluationContext
from yta_editor_parameters.values.abstract import VideoEditorParameterValue
from yta_editor_parameters.values import ConstantVideoEditorParameterValue
from yta_validation.parameter import ParameterValidator
from typing import Union
from dataclasses import dataclass


# TODO: We need to define all the specific effect
# params to be able to build them properly missing
# no fields
@dataclass
class WavingFrameVideoEffectParameters(EffectParameters):
    """
    The parameters to use when applying this effect
    that turns the frames into waving frames for a
    specific `evaluation_context`.

    This will be returned by the effect when
    calculated, for a specific `evaluation_context`.
    """

    def __init__(
        self,
        amplitude: float,
        frequency: float,
        speed: float
    ):
        # TODO: Add docs
        self.amplitude: float = amplitude
        self.frequency: float = frequency
        self.speed: float = speed
        

class WavingFrameVideoEffect(VideoEffect):
    """
    The effect that will modify the frames and
    transform them into waving frames according
    to the provided conditions.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        amplitude: VideoEditorParameterValue = ConstantVideoEditorParameterValue(0.05),
        frequency: VideoEditorParameterValue = ConstantVideoEditorParameterValue(10.0),
        speed: VideoEditorParameterValue = ConstantVideoEditorParameterValue(2.0)
    ):
        ParameterValidator.validate_mandatory_instance_of('amplitude', amplitude, VideoEditorParameterValue)
        ParameterValidator.validate_mandatory_instance_of('frequency', frequency, VideoEditorParameterValue)
        ParameterValidator.validate_mandatory_instance_of('speed', speed, VideoEditorParameterValue)

        self.do_use_gpu: bool = do_use_gpu
        """
        Flag to indicate if using GPU or not.
        """
        # TODO: Maybe 'parse_params' (?)
        self.amplitude: VideoEditorParameterValue = amplitude
        """
        The `VideoEditorParameterValue` that defines the value that
        should be applied for the `amplitude` and the specific
        `evaluation_context` provided.
        """
        self.frequency: VideoEditorParameterValue = frequency
        """
        The `VideoEditorParameterValue` that defines the value that
        should be applied for the `frequency` and the specific
        `evaluation_context` provided.
        """
        self.speed: VideoEditorParameterValue = speed
        """
        The `VideoEditorParameterValue` that defines the value that
        should be applied for the `speed` and the specific
        `evaluation_context` provided.
        """

    def _get_params_at(
        self,
        evaluation_context: EvaluationContext
    ) -> WavingFrameVideoEffectParameters:
        """
        *For internal use only*

        Get the parameters that must be applied at the given
        `evaluation_context`.
        """
        ParameterValidator.validate_mandatory_instance_of('evaluation_context', evaluation_context, EvaluationContext)

        return WavingFrameVideoEffectParameters(
            amplitude = self.amplitude.evaluate(
                evaluation_context = evaluation_context
            ),
            frequency = self.frequency.evaluate(
                evaluation_context = evaluation_context
            ),
            speed = self.speed.evaluate(
                evaluation_context = evaluation_context
            )
        )
    
    def _apply(
        self,
        # TODO: Set the type
        frame: any,
        output_size: Union[tuple[int, int], None],
        evaluation_context: EvaluationContext
    # TODO: Set the type
    ) -> any:
        parameters = self._get_params_at(
            evaluation_context = evaluation_context
        )

        # TODO: Is there any value that doesn't make any change (?)
        # if parameters.transparency == 1.0:
        #     return frame
        
        return WavingFrameVideoNodeProcessor(
            # TODO: What do I do with the 'opengl_context' (?)
            opengl_context = None
        ).process(
            inputs = {
                'base_input': frame
            },
            evaluation_context = evaluation_context,
            do_use_gpu = self.do_use_gpu,
            output_size = output_size,
            # parameters = parameters
            parameters = {
                'amplitude': parameters.amplitude,
                'frequency': parameters.frequency,
                'speed': parameters.speed
            }
        )


"""
TODO: Maybe I have to move the content of this.
"""
# from yta_editor.tracks.items.abstract import TrackItemWithAudioMedia, _TrackItemWithVideo
from yta_editor.tracks.items.transition.mode import TransitionMode
from yta_editor.tracks.items.video import _VideoTrackItem
from yta_editor.tracks.items.abstract import _TrackItemWithVideo, _TrackItemWithAudio
from yta_editor_time.evaluation_context import EvaluationContext
from quicktions import Fraction
from typing import Union


class TransitionTrackItem(_TrackItemWithVideo, _TrackItemWithAudio):
    """
    A track item that is built by joining 2 different video
    items that are consecutive.
    """

    @property
    def copy(
        self
    ) -> '_TrackItem':
        """
        A copy of the current instance.
        """
        return TransitionTrackItem(
            track = self._track,
            t_start = self.t_start,
            duration = self.duration,
            transition_specification = self._transition_specification,
            mode = self.mode,
            item_in = self.item_in,
            item_out = self.item_out
        )
    
    @property
    def fps(
        self
    ) -> float:
        """
        The `fps` associated to the `track` this transition
        item belongs to.
        """
        return self._track.fps
    
    @property
    def output_size(
        self
    ) -> tuple[int, int]:
        """
        The `output_size` of the `track` that is associated to
        this transition item.
        """
        return self._track.output_size
    
    @property
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the source has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.

        (!) This property is useless here as it is a
        transition.
        """
        return True

    def __init__(
        self,
        track: 'VideoTrack',
        t_start: Union[int, float, Fraction],
        # TODO: We need to be able to handle options
        # different than the 50% of each clip
        duration: Union[int, float, Fraction],
        transition_specification: Union['TransitionSpecification', None] = None,
        mode: TransitionMode = TransitionMode.TRIM,
        # TODO: This is a special one so this should be different
        item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
    ):
        mode = TransitionMode.to_enum(mode)

        if transition_specification is None:
            # TODO: Move to top if possible
            from yta_editor.tracks.items.transition.transition_specification import TransitionSpecification

            transition_specification = TransitionSpecification.default()

        super().__init__(
            track = track,
            t_start = t_start,
            duration = duration,
            item_in = item_in,
            item_out = item_out,
            # TODO: I think I don't need these transition fields
            # in this kind of item
            transition_in = None,
            transition_out = None
        )

        # I keep this by now to be able to copy it
        self._transition_specification: TransitionSpecification = transition_specification
        """
        The specification of the node and the parameters we
        want to be applied as the transition.
        """
        self.mode: TransitionMode = mode
        """
        The mode in which the transition must be implemented.
        """

        # TODO: Due to cyclic import issue
        # TODO: Move to top if possible
        from yta_editor.utils.transition_calculator import _TransitionCalculator
        
        self._transition_calculator: _TransitionCalculator = _TransitionCalculator(
            transition_item = self,
            duration = self.duration,
            # TODO: What about this 'do_use_gpu' (?)
            do_use_gpu = True,
            transition_specification = transition_specification
        )
        """
        *For internal use only*

        A transition calculator to simplify the way we handle
        the calculations to obtain the frames we need.
        """

    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        # TODO: Ignore this, we don't need it...
        do_use_source_limits: bool = True
    ) -> Union['AudioFrame', None]:
        """
        Get the audio frames of the provided `t_track`
        global time moment according to the internal
        configuration of the track item.

        TODO: This method must be overwritten by the
        specific class implementations.
        """
        # TODO: Is this the correct way? Is it generating
        # audio frames for a whole video frame (?)
        item_in_audio_frames = (
            list(self.item_in.get_audio_frames_at(
                t_track = t_track,
                # TODO: We should not receive the param
                do_use_source_limits = True
            ))
            if self.mode not in [TransitionMode.FREEZE_HEAD, TransitionMode.FREEZE_BOTH] else
            self._track.audio_silent
        )

        item_out_audio_frames = (
            list(self.item_out.get_audio_frames_at(
                # t_track = t_track + self.duration,
                t_track = t_track,
                # TODO: We should not receive the param
                do_use_source_limits = True
            ))
            if self.mode not in [TransitionMode.FREEZE_TAIL, TransitionMode.FREEZE_BOTH] else
            self._track.audio_silent
        )

        audio_frames = self._transition_calculator._get_process_audio_frames(
            audio_frames_a = item_in_audio_frames,
            audio_frames_b = item_out_audio_frames,
            evaluation_context = EvaluationContext.from_t(
                fps = self.fps,
                total_frames = int(self.duration * self.fps),
                t = self._t_track_to_t_track_item(t_track)
            )
        )

        # # TODO: Implement this as a combination of the audio
        # # coming from clips at the same time, but by now I'm
        # # just returning something
        # frames = [
        #     AudioFrameCombinator.sum_tracks_frames(
        #         tracks_frames = non_empty_collapsed_frames,
        #         sample_rate = self.audio_fps,
        #         # TODO: This was not being sent before
        #         layout = self.audio_layout,
        #         format = self.audio_format
        #     )
        # ]

        # for audio_frame in frames:
        #     yield audio_frame

        return audio_frames

        return item_in_audio_frames

    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Union['VideoFrame', None]:
        """
        Get the frame of the transition for the global
        `t_track` time moment provided.
        """
        # TODO: The 't' must be the global value of the timeline
        # Obtain the frames from each video
        # TODO: I actually need to obtain a frame that will be
        # out of the current 'item_in' valid time interval, as
        # it has been trimmed because of this transition, and 
        # the part we are using is just after (for the
        # 'item_in', before for the 'item_out') the last part
        # we are showing in the video
        # TODO: Calculate the 't' based on the 'item_in' t_end
        # TODO: We need to read the 'item_in.t_end' + 'local_t'
        # for the transition

        """
        Imagine this context:
        - item_in: [0, 2)   Transition: [2, 3)   item_out: [3, 4)
        The global t=2.5 is actually the 0.5 inside the 
        transition, that should read the 'item_in.t_end+0.5' (out
        of the time interval that is being shown) of the item_in.

        Now imagine that the item_in is [0, 2) but the media 
        inside is actually [0, 8) and it is reading from [3, 5).
        That means that the transition for the global t=2.5, 
        is a local_t=0.5, should be reading the t=2.5 (out of
        time interval), that is actually the t_media=5.5.
        """

        """
        As the media has been enshortened due to the transition,
        we need to use `t_track` values that are out of the 
        current item media limits, thats why we force the values
        to be allowed with the `do_use_source_limits` parameter.

        But, as the transition could be based on frozen frames,
        we 
        """
        frame_item_in = (
            self.item_in.get_video_frame_at(
                t_track = t_track,
                do_use_source_limits = True
            )
            if self.mode not in [TransitionMode.FREEZE_HEAD, TransitionMode.FREEZE_BOTH] else
            # TODO: Return as the one on top is doing
            self.item_in._last_video_frame
        ).as_rgba_numpy

        frame_item_out = (
            self.item_out.get_video_frame_at(
                #t_track = t_track + self.duration,
                t_track = t_track,
                do_use_source_limits = True
            )
            if self.mode not in [TransitionMode.FREEZE_TAIL, TransitionMode.FREEZE_BOTH] else
            # TODO: Return as the one on top is doing
            self.item_out._first_video_frame
        ).as_rgba_numpy

        # TODO: I think I need to implement the mask here
        # by using the 'self.item_mask.get_video_frame_at'

        """
        *Representacion con t(global) de clips*
        Clip A: [0, 3)   Clip B: [3, 5)
        - Añadimos transition de 1s
        Clip A: [0, 2)   Transition: [2, 3)   Clip B: [3, 4)
        - La 'Transition' tiene [2, 3) de A y [0, 1) de B
        - El t(global)=2.5 ahora cae en la transición, que se
          convierte a t(local)=2.5-(3.0-1.0)=0.5, y dado que
          la duration es 1.0; 0.5 es el t_progress=0.5
        """
        frame = self._transition_calculator.process_frame(
            frame_a = frame_item_in,
            frame_b = frame_item_out,
            evaluation_context = EvaluationContext.from_t(
                fps = self.fps,
                total_frames = int(self.duration * self.fps),
                t = self._t_track_to_t_track_item(t_track)
            )
        )

        # TODO: Improve this
        from yta_editor.utils.frame_wrapper import VideoFrameWrapped

        return VideoFrameWrapped(frame)
    
    # TODO: What if I have another abstract class to inherit
    # from that is different having not the transitions 
    # instead of blocking these methods here? Maybe a 
    # TrackItem, TrackItemWithTransitions (?)
    #  TODO: Should I remove this (?)
    def set_transition_in(self, transition):
        raise Exception('This class can not have transitions')
    
    def set_transition_out(self, transition):
        raise Exception('This class can not have transitions')
    
    def unset_transition_in(self):
        raise Exception('This class can not have transitions')
    
    def unset_transition_out(self):
        raise Exception('This class can not have transitions')  



"""
Note for the developers:
- All the media have, at least, audio, so we put
the audio effects array as a common one.
"""
from yta_editor_nodes.timeline.utils import validate_is_edition_node
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from yta_time_interval import TimeInterval
from quicktions import Fraction
from typing import Union
from abc import ABC, abstractmethod


class _Media(ABC):
    """
    Abstract class to be inherited by any media element.

    The media element is an element that includes a source
    and a pair of `t_start` and `t_end` values to be able to
    subclip that media source and use only the part we are
    interested in.

    A media will be always limited to the source duration,
    and the `t_start` and `t_end` time moments will be also
    limited to `0` and the duration of the source.
    """

    @property
    @abstractmethod
    def copy(
        self
    ) -> '_Media':
        """
        Get a copy of this instance with the same
        source, time range and effects.
        """
        pass

    @property
    def is_unchanged(
        self
    ) -> bool:
        """
        *For internal use only*

        Internal boolean flag to indicate if the media has
        been modified (or not), useful when we want to 
        optimize the way we access to the frames to render.

        If the original source has not been modified, we
        don't need to read frame by frame and apply any
        change so we can do it faster.

        (!) This property has to be modified by the specific
        video medias.
        """
        return (
            self.source.is_unchanged and
            self.head == 0 and
            self.tail == 0 and
            len(self._audio_effects) == 0
        )

    @property
    def head(
        self
    ) -> Fraction:
        """
        The time remaining at the begining of this
        Media source, which is the value between `0`
        and the `t_start` time moment.

        This value can be useful for transitions.

        The formula:
        - `self._time_interval.head`
        """
        return self._time_interval.head
    
    @property
    def tail(
        self
    ) -> Fraction:
        """
        The time remaining at the t_end of this Media
        source, which is the value between the `t_end`
        time moment and the total duration of the
        source.

        This value can be useful for transitions.

        The formula:
        - `self._time_interval.tail`
        """
        return self._time_interval.tail
    
    @property
    def t_start(
        self
    ) -> Fraction:
        """
        The time moment of the source in which this media
        should t_start being played.
        """
        return self._time_interval.t_start
    
    @property
    def t_end(
        self
    ) -> Fraction:
        """
        The time moment of the source in which this media
        should stop being played.
        """
        return self._time_interval.t_end
    
    @property
    def start_limit(
        self
    ) -> Fraction:
        """
        The lower time moment value that can be used to get
        the frame from the source. This value can be reached
        only if some boolean parameters are provided to force
        this limit, or the normal `[0, self.duration]` time
        interval should be used as the limit instead.

        If the media is pointing to the whole source duration,
        this limit will be `0`.
        """
        return -self.head

    @property
    def end_limit(
        self
    ) -> Fraction:
        """
        The upper time moment value that can be used to get
        the frame from the source. This value can be reached
        only if some boolean parameters are provided to force
        this limit, or the normal `[0, self.duration]` time
        interval should be used as the limit instead.

        If the media is pointing to the whole source duration,
        this limit will be `self.duration`.
        """
        return self.duration + self.tail
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the media in this specific moment
        as it has been defined by the user (that could
        have modified it) according to the current `t_start`
        and `t_end` values.

        The formula:
        - `self._time_interval.duration`
        """
        return self._time_interval.duration
    
    @property
    def max_duration(
        self
    ) -> Fraction:
        """
        The maximum duration of the media, that is the
        duration of the file source (if existing) or the
        one defined when instantiating it.

        The formula:
        - `self._time_interval.t_end_limit`
        """
        return self._time_interval.t_end_limit

    def __init__(
        self,
        source: Union['AudioFileSource', 'AudioNumpySource', 'VideoFileSource', 'VideoColorSource', 'VideoImageSource', 'VideoNumpySource'],
        t_start: Union[int, float, Fraction] = 0.0,
        t_end: Union[int, float, Fraction, None] = None,
        audio_effects: list[Union['SerialNode', 'ParallelNode']] = [],
    ):
        self.source: Union['AudioFileSource', 'AudioNumpySource', 'VideoFileSource', 'VideoColorSource', 'VideoImageSource', 'VideoNumpySource'] = source
        """
        The source of this media element that is the entity
        from which we can obtain the frames.
        """
        self._audio_effects: list['SerialNode', 'ParallelNode'] = []
        """
        The audio effects we want to apply on the media.
        """
        if (
            not PythonValidator.is_instance_of(source, ['AudioFileSource', 'VideoFileSource']) and
            t_end is None
        ):
            raise Exception('The `t_end` parameter is mandatory when the source is not a video/audio file.')
        
        # TODO: Here we don't have `fps` to be able to use them,
        # but we will limit with all the limits as we know how
        # long can the media be
        end_limit = (
            self.source.duration
            if self.source.duration is not None else
            # TODO: What do we do with this? It can't be None...
            None
        )

        self._time_interval: TimeInterval = TimeInterval(
            t_start = t_start,
            t_end = t_end,
            t_start_limit = 0,
            t_end_limit = end_limit,
            duration_limit = end_limit,
            fps = None
        )
        """
        *For internal use only*

        Internal time interval to handle the duration and be
        able to limit it to the real source duration.
        """
        
        self.set_audio_effects(audio_effects)

    def shift_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Update the `t_start` value by adding the `delta` provided
        (that can be positive to make it start later, or negative
        to make it start earlier), and also the `t_end` the same
        amount.

        This is useless when the media is pointing to the time
        interval `[2, 4)` of the source, for example, and we want
        to make it point to the `[3, 5)`. We can use a `shift_by(1)`
        in this case to achieve this purpose.
        """
        self._time_interval.shift_by(delta)

        return self
    
    def shift_to(
        self,
        t: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Update the `t_start` value to the time moment provided as
        `t`, and also the `t_end` the same amount of time and in
        the same direction).

        This is useful when the media is pointing to the time
        interval `[2, 4)` of the source, for example, and we want
        to make it point to the `[3, 5)`. We can use a `shift_to(3)`
        in this case to achieve this purpose.
        """
        self._time_interval.shift_to(t)

        return self

    def shift_start_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Update the `t_start` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        self._time_interval.shift_start_by(delta)

        return self
    
    def shift_start_to(
        self,
        t: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Update the `t_start` value to the `t` time moment
        provided (that can be greater or smaller than the
        current `t_start`), if possible and valid.
        """
        self._time_interval.shift_start_to(t)

        return self
    
    def shift_end_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Update the `t_end` value by adding the `delta` provided
        (that can be positive to make it end later or negative
        to make it end earlier).
        """
        self._time_interval.shift_end_by(delta)

        return self
    
    def shift_end_to(
        self,
        t: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Update the `t_end` value to the `t` time moment
        provided (that can be greater or smaller than the
        current `t_end`), if possible and valid.
        """
        self._time_interval.shift_end_to(t)

        return self
    
    def split(
        self,
        t: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Split the media by the `t` time moment provided
        and return the 2 new instances.
        """
        new_time_intervals = self._time_interval.split(t)

        left = self.copy
        right = self.copy

        left._time_interval.with_start_and_end(new_time_intervals[0].t_start, new_time_intervals[0].t_end)
        right._time_interval.with_start_and_end(new_time_intervals[1].t_start, new_time_intervals[1].t_end)

        return (
            left,
            right
        )
    
    def cut(
        self,
        t_start: Union[int, float, Fraction],
        t_end: Union[int, float, Fraction]
    ) -> '_Media':
        """
        Cut the media by the `t_start` and `t_end` time moments
        provided and get the new instance.
        """
        self._time_interval.cut(t_start, t_end)

        return self
    
    def add_audio_effect(
        self,
        effect: Union['SerialNode', 'ParallelNode']
    ) -> 'VideoFileMedia':
        """
        Add the provided audio `effect` to the list to apply
        to this media.
        """
        # TODO: Validate that it is actually an audio effect
        # inside the `effect`
        validate_is_edition_node(effect)

        self._audio_effects.append(effect)

        return self
    
    def set_audio_effects(
        self,
        effects: list[Union['SerialNode', 'ParallelNode']]
    ) -> 'VideoFileMedia':
        """
        Set the provided audio `effects` as the effects to
        apply to this media, replacing the previous ones.
        """
        # TODO: Validate that it is actually a list of audio
        # effects inside the `effects`
        self._audio_effects = effects

        return self
    
    # TODO: This method has been created
    # to be inherited by the other classes
    # and being able to copy the instance
    # properly by using the same 'source'
    # reference and creating not a new one
    @classmethod
    def _init_with_source(
        cls,
        source: Union['AudioFileSource', 'AudioNumpySource', 'VideoFileSource', 'VideoColorSource', 'VideoImageSource', 'VideoNumpySource'],
        t_start: Union[int, float, Fraction] = 0.0,
        t_end: Union[int, float, Fraction, None] = None,
        audio_effects: list[Union['SerialNode', 'ParallelNode']] = [],
    ):
        """
        *For internal use only*

        Alternative '__init__' to create the
        instance from the 'source' directly. This
        method must be called by the specific
        implementations of this abstract class
        to be able to instantiate them directly
        from the 'source' to make copies.

        We created this method to avoid generating
        a new 'source' instance but preserving the
        same reference.
        """
        # Create new instance skipping '__init__'
        instance = cls.__new__(cls)
        super(cls, instance).__init__(
            source = source,
            t_start = t_start,
            t_end = t_end,
            audio_effects = audio_effects
        )

        return instance
    
    def _t_media_to_t_source(
        self,
        t_media: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Fraction:
        """
        Transform the `t_media` received to the corresponding
        `t_source` according to the media duration.

        Imagine a source that lasts `8s`, and a media that is
        pointing to the time interval `[2, 4)` of that source,
        that is a time interval `[0, 2)` because of the 2s of
        duration. If they ask for the `t_media=1` (which is
        valid because it is in the `[0, 2)` valid interval),
        they are really asking for the `t_source=3` 
        (the `t_media` plus the `media.t_start`).

        The `do_use_source_limits` boolean parameter will allow
        us to use values beyond the media limits (to be able to
        obtain frames from the source that could be useful when
        a transition is being built).
        """
        start_limit = (
            self.start_limit
            if do_use_source_limits else
            0
        )

        end_limit = (
            self.end_limit
            if do_use_source_limits else
            self.duration
        )

        """
        Source [0, 8]
        Media  [2, 4] => [0, 2]
        t_media = -3 not in [0, 2] => -1 not in [0, 8]
        t_media = -2 not in [0, 2] => 0 in [0, 8]
        t_media = -1 not in [0, 2] => 1 in [0, 8]
        t_media = 0 in [0, 2] => 2 in [0, 8]
        t_media = 1 in [0, 2] => 3 in [0, 8]
        t_media = 2 in [0, 2] => 4 in [0, 8]
        t_media = 3 not in [0, 2] => 5 in [0, 8]
        t_media = 6 not in [0, 2] => 8 in [0, 8]
        t_media = 7 not in [0, 2] => 9 not in [0, 8]
        """
        ParameterValidator.validate_mandatory_number_between('t_media', t_media, start_limit, end_limit, True, False)

        return t_media + self.t_start
    
    def __str__(
        self
    ) -> str:
        """
        Method to return the information about this media as
        a printable str.
        """
        return f'TimeInterval: [{self.t_start}, {self.t_end}). Limits: [{self.start_limit}, {self.end_limit}).'
    
"""
If the source has a duration of 8s, it means that
the time interval is [0, 8). The media will have
a maximum duration of 8s and the limits will be
those values (0 and 8). The media can be set to
[2, 4), which means that will be reading that time
interval of the source, but will be [0, 2) 
internally, so you should ask for t_media=1 to
obtain the media value, which will be transformed
into the t_source=3. Considering these, you should
be able to ask for t_media=-1, that would be
transformed into the t_source=1 that is valid.

The media will be placed (contained) in an item of
the track, that will be set in a time interval of
that track. So, the t_track=8, if that item (that
includes that media) is set in the track time
interval [7, 9), asking for the t_track=8 will be
transformed (by the item) into ask for the
t_media=1, that will be also transformed (by the
media) into the t_source=3.
"""